package me.totalfreedom.discord;

import discord4j.common.util.Snowflake;
import discord4j.core.DiscordClientBuilder;
import discord4j.core.GatewayDiscordClient;
import discord4j.core.event.domain.interaction.ChatInputInteractionEvent;
import discord4j.core.object.entity.Guild;
import discord4j.core.object.entity.Member;
import discord4j.core.object.entity.Role;
import discord4j.core.object.entity.channel.TextChannel;
import me.totalfreedom.discord.command.HelpCommand;
import me.totalfreedom.discord.command.ListCommand;
import me.totalfreedom.discord.command.TPSCommand;
import me.totalfreedom.discord.handling.CommandHandler;
import me.totalfreedom.discord.util.SnowflakeEntry;
import me.totalfreedom.discord.util.TFM_Bridge;
import me.totalfreedom.totalfreedommod.admin.Admin;
import me.totalfreedom.totalfreedommod.config.ConfigEntry;
import me.totalfreedom.totalfreedommod.player.PlayerData;
import me.totalfreedom.totalfreedommod.rank.Rank;
import me.totalfreedom.totalfreedommod.util.FLog;
import net.dv8tion.jda.internal.utils.concurrent.CountingThreadFactory;
import org.apache.commons.lang.RandomStringUtils;
import org.bukkit.Bukkit;
import reactor.core.publisher.Mono;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ScheduledThreadPoolExecutor;

public class Bot
{
    private final GatewayDiscordClient client;
    private final TFM_Bridge tfm;
    private final HashMap<String, PlayerData> LINK_CODES = new HashMap<>();
    private ScheduledThreadPoolExecutor RATELIMIT_EXECUTOR;
    private Boolean enabled = false;


    public Bot()
    {
        //Creates the gateway client and connects to the gateway
        this.client = DiscordClientBuilder.create(ConfigEntry.DISCORD_TOKEN.getString())
                .build()
                .login()
                .block();

        this.tfm = new TFM_Bridge(this);
    }

    public void initialize() {
        if (client == null) throw new IllegalStateException();

        final CommandHandler handler = new CommandHandler(client.getRestClient());

        /* Call our code to handle creating/deleting/editing our global slash commands.
        We have to hard code our list of command files since iterating over a list of files in a resource directory
         is overly complicated for such a simple demo and requires handling for both IDE and .jar packaging.
         Using SpringBoot we can avoid all of this and use their resource pattern matcher to do this for us.
         */
        List<String> commands = List.of("greet.json", "ping.json");

        try
        {
            handler.registerCommands(commands);
            handler.registerCommand(new HelpCommand());
            handler.registerCommand(new ListCommand());
            handler.registerCommand(new TPSCommand());
        } catch (Exception e)
        {
            Bukkit.getLogger().severe("Error trying to register global slash commands.\n" + e.getMessage());
        }
        //Register our slash command listener
        client.on(ChatInputInteractionEvent.class, handler::handle)
                .then(client.onDisconnect())
                .block(); // We use .block() as there is not another non-daemon thread and the jvm would close otherwise.

        RATELIMIT_EXECUTOR = new ScheduledThreadPoolExecutor(5, new CountingThreadFactory(this::poolIdentifier, "RateLimit"));
        RATELIMIT_EXECUTOR.setRemoveOnCancelPolicy(true);
    }

    public String formatBotTag() {
        return client.getSelf()
                .blockOptional()
                .orElseThrow()
                .getUsername() + "#" + client.getSelf()
                .blockOptional()
                .orElseThrow()
                .getDiscriminator();
    }


    private String poolIdentifier()
    {
        return "TFD4J";
    }

    public TFM_Bridge getTFM()
    {
        return tfm;
    }

    public Mono<Guild> getGuildById()
    {
        return client.getGuildById(SnowflakeEntry.serverID);
    }

    public GatewayDiscordClient getClient()
    {
        return client;
    }

    public Map<String, PlayerData> getLinkCodes()
    {
        return LINK_CODES;
    }

    public boolean shouldISendReport()
    {
        if (ConfigEntry.DISCORD_REPORT_CHANNEL_ID.getString().isEmpty())
        {
            return false;
        }

        if (ConfigEntry.DISCORD_SERVER_ID.getString().isEmpty())
        {
            FLog.severe("No Discord server ID was specified in the config, but there is a report channel ID.");
            return false;
        }

        Guild server = client.getGuildById(SnowflakeEntry.serverID).block();
        if (server == null)

        {
            FLog.severe("The Discord server ID specified is invalid, or the bot is not on the server.");
            return false;
        }

        return true;
    }
}
